package de.hsbo.fbv.bmg.geometry.simple2hd;

import java.util.ArrayList;
import java.util.List;

import de.hsbo.fbv.bmg.geometry.simple.CoordinateM;
import de.hsbo.fbv.bmg.geometry.simple.GPoint;
import de.hsbo.fbv.bmg.geometry.simple.GTriangle;
import de.hsbo.fbv.bmg.geometry.simple.GeoModel;
import de.hsbo.fbv.bmg.geometry.simple.collections.GPointArray;
import de.hsbo.fbv.bmg.geometry.simple.collections.GPointSet;



/**
 * Die Klasse stellt die notwendigen Algorithmen fr eine Dreiecksvermaschung 
 * (Delaunay-Triangulation) als statische Methoden bereit.<br>
 * Der Algorithmus basiert auf dem rekursiven Ansatz von Lee u. Schachter, 
 * (Lee, Schachter: Two Algorithms for Constructing a Delaunay Triangulation, 
 * International Journal of Computer and Information Sciences, Vol. 9, No. 3, 1980)
 * @author zimmermann
 *
 */
public class Delaunay {

	// Parameter
	public static final int minx = 0, W = 0, SW = 0;
	public static final int miny = 1, S = 1, SE = 1;
	public static final int maxx = 2, E = 2, NW = 2;
	public static final int maxy = 3, N = 3, NE = 3;	
	
	public static GTriangle lastT = null;
	public static GPoint lastP = null;
	
	/**
	 * Algorithmus zum Aufbau eines TINs aus der Punktmenge pset.<br>
	 * @param pset
	 * @return
	 */
	public static synchronized List<GTriangle> getTIN(GPointSet pset) {
		
		List<GTriangle> tin = new ArrayList<GTriangle>();
		
		try {
			// 1. Teil: Initialisierung
			// Erzeuge ein Punktfeld pinit (Hilfsmanahme)
			GPointArray pinit = new GPointArray(pset.toArray());
			// Erzeuge einen Rechteckrahmen um die Punktmenge herum
			GPoint[] corner = pinit.getABox(1.1);
			// Bestimme den Lageschwerpunkt und suche den dazu nchsten Punkt
			GPoint center = pinit.nearestPoint(pinit.getCenter());
			
			// Bilde die vier ersten Dreiecke und ihre Nachbarverweise
			GTriangle[] tinit = new GTriangle[4];
			tinit[N] = GeoModel.createTriangle(corner[NW], center, corner[NE]);
			tinit[W] = GeoModel.createTriangle(corner[SW], center, corner[NW]);
			tinit[S] = GeoModel.createTriangle(corner[SE], center, corner[SW]);
			tinit[E] = GeoModel.createTriangle(corner[NE], center, corner[SE]);
			tinit[N].setNextTriangles(tinit[W], tinit[E], null);
			tinit[W].setNextTriangles(tinit[N], tinit[S], null);
			tinit[S].setNextTriangles(tinit[W], tinit[E], null);
			tinit[E].setNextTriangles(tinit[N], tinit[S], null);
			
			// Trage die vier ersten Dreiecke in die Liste ein
			// List<GTriangle> tin = new ArrayList<GTriangle>();
			tin.add(tinit[N]);
			tin.add(tinit[W]);
			tin.add(tinit[S]);
			tin.add(tinit[E]);
			
			// Bestimme Extrempunkte und fge sie ins Maschennetz ein
			GPoint[] minMaxP = pinit.getExtremePointsXY();
			for(int i=0; i<minMaxP.length; i++)
				add(tin, minMaxP[i]);
			
			// Entferne den Lageschwerpunkt und die Extrempunkte aus der Punktmenge ptset
			pset.remove(center);
			for(int i=0; i<minMaxP.length; i++)
				pset.remove(minMaxP[i]);
			// Erzeuge ein neues Punktfeld
			pinit = new GPointArray(pset.toArray());
			// Zerlege das Punktfeld in vier neue Quadranten
			GPointArray[] pquads = pinit.splitXY(center.x(), center.y());
			// Gib Objekt pinit auf (wird nicht mehr gebraucht)
			pinit = null;
			
			// 2. Teil: Rekursion - Divide-And-Conquer-Algorithmus	
			// Fge den Lageschwerpunkt in die Dreiecksliste tin ein
			// Zerlege die Punktmenge am Lageschwerpunkt in vier neue Quadranten
			// Wiederhole solange, bis keine neuen Punkte mehr vorhanden sind
			for(int i=0; i<pquads.length; i++)
				if(pquads[i]!=null)
					insert(tin, pquads[i]);
					
			// 3. Teil: Nachbearbeitung
			// Lsche alle Dreiecke, die an den Corner-Punkten liegen
			
			// Ende: Dreieckliste tin zurckgeben
			return tin;
			
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Rekursiver Teil - Divide-And-Conquer-Algorithmus
	 * @param tin
	 * @param parray
	 */
	public static synchronized void insert(List<GTriangle> tin, GPointArray parray) throws Exception {
		// Wenn das Punktfeld mehr als vier Punkte enthlt ...
		if (parray.size()>4) {
			// Bestimme den Lageschwerpunkt und suche den dazu nchsten Punkt
			GPoint center = parray.nearestPoint(parray.getCenter());
			// Fge den Lageschwerpunkt ins TIN ein
			add(tin, center);
			// Entferne den Lageschwerpunkt aus dem Punktfeld
			GPointSet pset = new GPointSet(parray.toArray());
			pset.remove(center);
			parray = new GPointArray(pset.toArray());
			// Gib Objekt pset auf (wird nicht mehr gebraucht)
			pset = null;
			// Zerlege das Punktfeld in vier neue Quadranten
			GPointArray[] pquads = parray.splitXY(center.x(), center.y());
			// Gib Objekt parray auf (wird nicht mehr gebraucht)
			parray = null;
			// Rufe diesen Algorithmus rekursiv auf
			for (int i = 0; i < pquads.length; i++)
				if (pquads[i] != null)
					insert(tin, pquads[i]);
		}
		// Sonst ... (Punktfeld enthlt nur noch maximal vier Punkte)
		else {
			// Fge alle verbleibenden Punkt sequenziell ein
			for(int i=0; i<parray.size(); i++)
				add(tin, parray.get(i));			
			// Gib Objekt parray auf (wird nicht mehr gebraucht)
			parray = null;
		}
	}
		
		
	/**
	 * Fgt einen neuen Punkt ins Maschennetz ein.<br>
	 * @param list
	 * @param p
	 */
	public static synchronized void add(List<GTriangle> list, GPoint p) throws Exception {
		// Suche das passende Dreieck
		 GTriangle tri = findTriangle(list, p);
		// Baue die Dreiecksvermaschung am Punkt p um
		reconstruct(list, tri, p);
	}
	
	/**
	 * Die Methode zerlegt ein bestehendes Dreieck in drei neue, falls der Punkt p
	 * im Dreieck liegt. Anschlieend werden die benachbarten Dreiecke zerlegt, wenn 
	 * der Punkt in ihrem Umkreis liegt. Dadurch entstehen insgesamt drei bis sechs 
	 * neue Dreiecke.<br> 
	 * @param list Liste der bestehenden Dreiecke
	 * @param tri Index des Dreiecks, das zerlegt werden soll
	 * @param p Punkt (CoordinateM), der die neuen Dreiecke bildet.
	 */
	static synchronized void reconstruct(List<GTriangle> list, GTriangle tri,
			GPoint p) throws Exception {
		
		// Wenn der Punkt im Dreieck liegt, zerlege es in drei neue Dreiecke
		if (tri.contains(p)) {
			// Alle Koordinaten abrufen
			CoordinateM d = p.getCoordinate();
			CoordinateM a = tri.getPoint(GTriangle.A);
			CoordinateM b = tri.getPoint(GTriangle.B);
			CoordinateM c = tri.getPoint(GTriangle.C);
			// Generiere neue Dreiecke ABD, BCD und CAD
			GTriangle tABD = new GTriangle(a, b, d);
			GTriangle tBCD = new GTriangle(b, c, d);
			GTriangle tCAD = new GTriangle(c, a, d);
			// Benachbarte Dreiecke zuweisen - tABD
			GTriangle nABD = tri.getNextTriangle(GTriangle.AB);
			tABD.setNextTriangles(nABD, tBCD, tCAD);
			if (nABD != null)
				nABD.setNextTriangle(tABD);
			// Benachbarte Dreiecke zuweisen - tBCD
			GTriangle nBCD = tri.getNextTriangle(GTriangle.BC);
			tBCD.setNextTriangles(nBCD, tCAD, tABD);
			if (nBCD != null)
				nBCD.setNextTriangle(tBCD);
			// Benachbarte Dreiecke zuweisen - tCAD
			GTriangle nCAD = tri.getNextTriangle(GTriangle.CA);
			tCAD.setNextTriangles(nCAD, tABD, tBCD);
			if (nCAD != null)
				nCAD.setNextTriangle(tCAD);
			// Lsche Dreieck tri aus der Liste und fge die neuen Dreiecke ein
			list.remove(tri);
			// TODO NUR ZU TESTZWECKEN
			lastT = tri;
			lastP = p;
			list.add(tABD);
			list.add(tBCD);
			list.add(tCAD);
			// Fhre den Algorithmus erneut fr die Nachbar-Dreiecke durch
			if (nABD != null)
				reconstruct(list, nABD, tABD, p);
			if (nBCD != null)
				reconstruct(list, nBCD, tBCD, p);
			if (nCAD != null)
				reconstruct(list, nCAD, tCAD, p);
			
			// Ende der Methode
			return;
		}

		// Sonst keine Aktion
		else {
			throw new Exception("Punkt " + p + " liegt nicht im Dreieck " + tri);
		}
	}	
	
	/**
	 * Die Methode bildet aus zwei benachbarten Dreiecken zwei neue Dreiecke, 
	 * wenn der Punkt p im Umkreis des Dreiecks mit dem Index itri liegt. Andernfalls 
	 * bleiben beide Dreiecke unverndert bestehen.
	 * @param list Liste der bestehenden Dreiecke
	 * @param tri Dreieck, in dessen Umkreis der Punkt liegen soll
	 * @param nexttri benachbartes Dreieck, zu dem der Punkt p gehrt
	 * @param p Punkt (CoordinateM)
	 */
	static synchronized void reconstruct(List<GTriangle> list, GTriangle tri,
			GTriangle nexttri, GPoint p) throws Exception {
		// Wenn der Punkt im Umkreis des Dreiecks tri liegt, gestalte es um!
		if (tri.isInCircle(p.getCoordinate())) {

			List<CoordinateM> side = tri.getCommonPoints(nexttri);
			CoordinateM a = side.get(0);
			CoordinateM b = side.get(1);
			CoordinateM c = tri.getOppositePoint(a, b);
			CoordinateM d = p.getCoordinate();
			if (!d.equals2D(nexttri.getOppositePoint(a, b)))
				throw new Exception("points are differing!");
			// Generiere neue Dreiecke ACD und BDC
			GTriangle tACD = new GTriangle(a, c, d);
			GTriangle tBDC = new GTriangle(b, d, c);
			// Benachbarte Dreiecke abrufen
			GTriangle nACD1 = tri.getNextTriangle(tri.indexOf(a, c));
			GTriangle nACD2 = nexttri.getNextTriangle(nexttri.indexOf(a, d));
			GTriangle nBDC1 = tri.getNextTriangle(tri.indexOf(b, c));
			GTriangle nBDC2 = nexttri.getNextTriangle(nexttri.indexOf(b, d));
			// Benachbarte Dreiecke neu zuordnen
			tACD.setNextTriangles(tBDC, nACD1, nACD2);
			tBDC.setNextTriangles(tACD, nBDC1, nBDC2);
			// Gegenbeziehungen der benachbarten Dreiecke auch neu zuordnen
			if(nACD1!=null)
				nACD1.setNextTriangle(tACD);
			if(nACD2!=null)
				nACD2.setNextTriangle(tACD);
			if(nBDC1!=null)
				nBDC1.setNextTriangle(tBDC);
			if(nBDC2!=null)
				nBDC2.setNextTriangle(tBDC);			
			// Lsche Dreiecke tri, nexttri aus der Liste und fge die neuen
			// Dreiecke ein
			list.remove(tri);
			list.remove(nexttri);
			list.add(tACD);
			list.add(tBDC);
			
			// Erneut rekursiv aufrufen fr deren benachbarte Dreiecke!
			if(nACD1!=null && nACD1.indexOf(d)==-1) 
				reconstruct(list, nACD1, tACD, p);
			if(nACD2!=null && nACD2.indexOf(d)==-1) 
				reconstruct(list, nACD2, tACD, p);
			if(nBDC1!=null && nBDC1.indexOf(d)==-1) 
				reconstruct(list, nBDC1, tBDC, p);
			if(nBDC2!=null && nBDC2.indexOf(d)==-1)
				reconstruct(list, nBDC2, tBDC, p);
			return;
		}

		// Sonst keine Aktion
		else
			return;
	}
	
	/**
	 * Einfacher Algorithmus zur Suche nach dem Dreieck, in dem der Punkt p liegt.<br>
	 * Der Algorithmus besitzt lineares Wachstum O(n) !
	 * @param list
	 * @param p
	 * @return
	 */
	static synchronized GTriangle findTriangle(List<GTriangle> list, GPoint p) throws Exception {
		for(int i=0; i<list.size(); i++)
			if(((GTriangle)list.get(i)).contains(p))
				return (GTriangle)(list.get(i));
		return null;
	}

}
